###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=protected-access
from concurrent.futures import ThreadPoolExecutor
from os import environ
from time import sleep
from uuid import uuid4
from redis import exceptions, client # pylint: disable=import-self

from everysk.config import settings, SettingsManager
from everysk.core import redis as redis_module
from everysk.core.redis import (
    RedisCache, RedisCacheCompressed, RedisChannel, RedisClient,
    RedisEmptyListError, RedisList, RedisLock, cache
)
from everysk.core.unittests import TestCase, mock


@cache(timeout=1)
def sum(a: int, b: int = 0) -> int: # pylint: disable=redefined-builtin
    return a + b


class CacheDecoratorTestCase(TestCase):

    def setUp(self) -> None:
        self.obj = RedisCache()
        self.obj.flush_all()

    def test_timeout_value_not_int(self):
        with self.assertRaises(ValueError) as context:
            cache(timeout='1')

        self.assertEqual(str(context.exception), 'Timeout must be an integer greater than 0.')

    def test_timeout_value_zero(self):
        with self.assertRaises(ValueError) as context:
            cache(timeout=0)

        self.assertEqual(str(context.exception), 'Timeout must be an integer greater than 0.')

    def test_timeout_value_negative(self):
        with self.assertRaises(ValueError) as context:
            cache(timeout=-1)

        self.assertEqual(str(context.exception), 'Timeout must be an integer greater than 0.')

    def test_cache_decorator(self):
        self.assertEqual(sum(1, 1), 2)
        self.assertEqual(sum(1, 1), 2)
        self.assertListEqual(
            self.obj.connection.keys(),
            [b'sum:e718fe8edf8a7c2917cd7444286b7048042a74abf5208f1d70933c08dbd6b7e9']
        )
        self.assertDictEqual(sum.info, {'hits': 1, 'misses': 1})
        sleep(1.1)
        self.assertEqual(sum(1, 1), 2)
        self.assertDictEqual(sum.info, {'hits': 1, 'misses': 2})
        self.assertEqual(sum(a=1, b=1), 2)
        self.assertDictEqual(sum.info, {'hits': 1, 'misses': 3})
        self.assertEqual(sum(1), 1)
        self.assertDictEqual(sum.info, {'hits': 1, 'misses': 4})

    def test_cache_decorator_clear(self):
        self.obj.set('key', 1)
        self.assertEqual(sum(1, 1), 2)
        self.assertEqual(sum(1, 2), 3)
        self.assertEqual(sum(1, 3), 4)
        self.assertListEqual(
            sorted(self.obj.connection.keys()),
            [
                b'2c70e12b7a0646f92279f427c7b38e7334d8e5389cff167a1dc30e73f826b683',
                b'sum:3a316d6d3226f84c1e46e4447fa8d5fd800bff4a1bc6498152523cd4a602b69b',
                b'sum:4e3970018fa500922ce5c087f84c734d7b47b0651da2f2e17226801885838032',
                b'sum:e718fe8edf8a7c2917cd7444286b7048042a74abf5208f1d70933c08dbd6b7e9',
            ]
        )
        sum.clear()
        self.assertListEqual(self.obj.connection.keys(), [b'2c70e12b7a0646f92279f427c7b38e7334d8e5389cff167a1dc30e73f826b683'])


class RedisClientTestCase(TestCase):

    def setUp(self) -> None:
        RedisClient._connection = {}

    @mock.patch.object(redis_module, 'log')
    def test_environ(self, log: mock.MagicMock):
        environ['REDIS_HOST'] = '0.1.0.1'
        environ['REDIS_PORT'] = '1234'
        redis = RedisClient()
        self.assertEqual(redis.host, environ['REDIS_HOST'])
        self.assertEqual(redis.port, int(environ['REDIS_PORT']))
        del environ['REDIS_HOST']
        del environ['REDIS_PORT']
        log.debug.assert_called_once_with('Connecting on Redis(%s:%s).....', redis.host, redis.port)

    @mock.patch.object(redis_module, 'log')
    def test_connection_property(self, log: mock.MagicMock):
        redis = RedisClient()
        self.assertEqual(redis.connection, redis._connection[redis._connection_key()])
        self.assertEqual(redis.connection, RedisClient._connection[redis._connection_key()])
        log.debug.assert_called_once()

    @mock.patch.object(redis_module, 'log')
    def test_hash_key(self, log: mock.MagicMock):
        redis = RedisClient()
        self.assertEqual(
            redis.get_hash_key('a'),
            'ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb'
        )
        # Check again to guarantee the same hash
        self.assertEqual(
            redis.get_hash_key('a'),
            'ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb'
        )
        log.debug.assert_called_once()

    @mock.patch.object(redis_module, 'log')
    def test_connect(self, log: mock.MagicMock):
        redis = RedisClient()
        log.debug.assert_called_once()
        RedisClient._connection = {}
        self.assertIsNone(RedisClient._connection.get(redis._connection_key(), None))
        redis._connect()
        self.assertIsNotNone(RedisClient._connection[redis._connection_key()])

    @mock.patch.object(RedisClient, '_connect')
    def test_connection_error_connection(self, _connect: mock.MagicMock):
        obj = RedisClient()
        redis_mock = mock.MagicMock(spec=redis_module.Redis)
        redis_mock.ping.side_effect = exceptions.ConnectionError
        with mock.patch.object(RedisClient, '_connection', {obj._connection_key(): redis_mock}) as _connection:
            connection = obj.connection # pylint: disable=unused-variable
        _connection[obj._connection_key()].ping.assert_called_once_with()
        _connect.assert_has_calls((mock.call(), mock.call()))

    @mock.patch.object(redis_module.log, 'debug', mock.MagicMock)
    def test_flush_all(self):
        cli = RedisClient()
        # Generate a random key to test
        key = uuid4().hex
        with SettingsManager(settings):
            settings.REDIS_SHOW_LOGS = True
            cli.connection.set(key, 'Flush-test')
            with mock.patch.object(redis_module.log, 'info') as debug:
                cli.flush_all()
            debug.assert_called_once_with('Redis flushed all keys.')
            self.assertIsNone(cli.connection.get(key))

            cli.connection.set(key, 'Flush-test')
            with mock.patch.object(redis_module.log, 'error') as debug:
                redis_mock = mock.MagicMock(spec=redis_module.Redis)
                redis_mock.flushall.return_value = 0
                with mock.patch.object(RedisClient, '_connection', {cli._connection_key(): redis_mock}) as _connect:
                    cli.flush_all()
            debug.assert_called_once_with('Redis flush all keys failed.')
            self.assertEqual(cli.connection.get(key), b'Flush-test')

    @mock.patch.object(redis_module.log, 'debug', mock.MagicMock)
    def test_build_prefix(self):
        cli = RedisClient(prefix='test')
        self.assertEqual(cli._build_prefix(), cli.prefix)
        self.assertEqual(cli._build_prefix('prefix'), 'prefix')

        with SettingsManager(settings):
            settings.REDIS_NAMESPACE = 'namespace'
            self.assertEqual(cli._build_prefix('prefix'), 'namespace:prefix')
            self.assertEqual(cli._build_prefix(), f'namespace:{cli.prefix}')
            settings.REDIS_NAMESPACE = None
            self.assertEqual(cli._build_prefix('prefix'), 'prefix')
            self.assertEqual(cli._build_prefix(), cli.prefix)

    @mock.patch.object(redis_module.log, 'debug', mock.MagicMock)
    def test_get_hash_key(self):
        cli = RedisClient()
        self.assertEqual(
            cli.get_hash_key('a'),
            'ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb'
        )
        # Check again to guarantee the same hash
        self.assertEqual(
            cli.get_hash_key('a'),
            'ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb'
        )
        self.assertRaisesRegex(
            ValueError,
            'Key cannot be None.',
            cli.get_hash_key,
            None
        )

@mock.patch.object(redis_module, 'log', mock.MagicMock())
class RedisCacheTestCase(TestCase):

    def setUp(self) -> None:
        RedisClient._connection = {}

    def tearDown(self):
        # Clean up the Redis cache after each test
        obj = RedisCache()
        obj.flush_all()

    def test_get_hash_key(self):
        obj = RedisCache()
        hash_key = '2c70e12b7a0646f92279f427c7b38e7334d8e5389cff167a1dc30e73f826b683'
        self.assertEqual(obj.get_hash_key('key'), hash_key)
        # Must always be the same value
        self.assertEqual(obj.get_hash_key('key'), hash_key)

    def test_get_set_delete(self):
        obj = RedisCache()
        obj.delete('key')
        self.assertIsNone(obj.get('key'))
        obj.set('key', 'My value.')
        self.assertEqual(obj.get('key'), b'My value.')
        obj.set('key', 1)
        self.assertEqual(obj.get('key'), b'1')
        obj.set('key', b'1.1')
        self.assertEqual(obj.get('key'), b'1.1')
        obj.delete('key')
        self.assertIsNone(obj.get('key'))

    def test_set_error(self):
        obj = RedisCache()
        self.assertRaises(exceptions.DataError, obj.set, 'key', ())
        self.assertRaises(exceptions.DataError, obj.set, 'key', [])
        self.assertRaises(exceptions.DataError, obj.set, 'key', {})
        self.assertRaises(exceptions.DataError, obj.set, 'key', obj)

    def test_prefix(self):
        key = b'prefix:2c70e12b7a0646f92279f427c7b38e7334d8e5389cff167a1dc30e73f826b683'
        obj = RedisCache(prefix='prefix')
        self.assertNotIn(key, obj.connection.keys())
        obj.set('key', 1)
        self.assertIn(key, obj.connection.keys())
        obj.delete('key')
        self.assertNotIn(key, obj.connection.keys())

    def test_delete_prefix(self):
        keys = [
            b'other:2c70e12b7a0646f92279f427c7b38e7334d8e5389cff167a1dc30e73f826b683',
            b'prefix:cd42404d52ad55ccfa9aca4adc828aa5800ad9d385a0671fbcbf724118320619',
            b'prefix:2c70e12b7a0646f92279f427c7b38e7334d8e5389cff167a1dc30e73f826b683',
        ]
        # Create a key with another prefix
        obj = RedisCache(prefix='other')
        obj.flush_all()
        obj.set('key', 1)

        obj = RedisCache(prefix='prefix')
        obj.set('key', 1)
        obj.set('value', 1)
        self.assertListEqual(sorted(obj.connection.keys()), sorted(keys))

        # Delete the key with other
        obj.delete_prefix(prefix='other')
        self.assertListEqual(sorted(obj.connection.keys()), sorted(keys[1:]))

        # Delete keys with prefix
        obj.delete_prefix()
        self.assertListEqual(obj.connection.keys(), [])

    def test_get_multi(self):
        obj = RedisCache()
        obj.set_multi({
            'key1': 1,
            'key2': 2,
            'key3': 3,
        })
        self.assertDictEqual(
            obj.get_multi(['key1', 'key2', 'key3']),
            {'key1': b'1', 'key2': b'2', 'key3': b'3'}
        )
        self.assertDictEqual(
            obj.get_multi(['key1', 'key2']),
            {'key1': b'1', 'key2': b'2'}
        )
        self.assertDictEqual(
            obj.get_multi(['key4']),
            {'key4': None}
        )

    def test_set_multi_error(self):
        obj = RedisCache()
        data = {
            'key1': 1,
            'key2': 2,
            'key3': 3,
        }
        redis_mock = mock.MagicMock(spec=redis_module.Redis)
        redis_mock.pipeline.return_value.execute.return_value = [True, True, False]
        with mock.patch.object(RedisClient, '_connection', {obj._connection_key(): redis_mock}) as _connection:
            with mock.patch.object(redis_module.log, 'error') as error:
                obj.set_multi(data)
            error.assert_called_once_with('Error RedisCache set_multi', extra={
                'labels': {
                    'REDIS_NAMESPACE': settings.REDIS_NAMESPACE,
                    'data': data,
                    'time': obj.timeout_default
            }})

    def test_delete_multi(self):
        obj = RedisCache()
        obj.set_multi({
            'key1': 1,
            'key2': 2,
            'key3': 3,
        })
        self.assertDictEqual(
            obj.get_multi(['key1', 'key2', 'key3']),
            {'key1': b'1', 'key2': b'2', 'key3': b'3'}
        )
        self.assertTrue(obj.delete_multi(['key1', 'key2']))
        self.assertDictEqual(
            obj.get_multi(['key1', 'key2']),
            {'key1': None, 'key2': None}
        )
        self.assertDictEqual(
            obj.get_multi(['key3']),
            {'key3': b'3'}
        )
        self.assertFalse(obj.delete_multi([]))
        self.assertTrue(obj.delete_multi('key3'))
        self.assertFalse(obj.delete_multi(['key3']))

    def test_incr(self):
        obj = RedisCache()
        self.assertEqual(obj.incr('key', initial_value=0), 1)
        self.assertEqual(obj.incr('key', initial_value=0), 2)
        self.assertEqual(obj.incr('key', delta=2, initial_value=0), 4)
        self.assertEqual(obj.incr('key', delta=2, initial_value=0), 6)
        self.assertEqual(obj.get('key'), b'6')

        self.assertRaisesRegex(
            ValueError,
            'Initial value must be set.',
            obj.incr,
            'key',
        )

        redis_mock = mock.MagicMock(spec=redis_module.Redis)
        redis_mock.pipeline.return_value.execute.return_value = [None, None]
        with mock.patch.object(RedisClient, '_connection', {obj._connection_key(): redis_mock}) as _connection:
            with mock.patch.object(redis_module.log, 'error') as error:
                obj.incr('key', initial_value=0)
        error.assert_called_once_with('Error RedisCache incr', extra={
            'labels': {
                'REDIS_NAMESPACE': settings.REDIS_NAMESPACE,
                'key': 'key',
                'delta': 1,
                'initial_value': 0,
            }
        })

    def test_decr(self):
        obj = RedisCache()
        self.assertEqual(obj.decr('key_1', initial_value=0), 0)
        self.assertEqual(obj.decr('key', initial_value=5), 4)
        self.assertEqual(obj.decr('key', delta=2, initial_value=0), 2)
        self.assertEqual(obj.decr('key', delta=2, initial_value=0), 0)
        self.assertEqual(obj.get('key'), b'0')

        self.assertRaisesRegex(
            ValueError,
            'Initial value must be set.',
            obj.decr,
            'key',
        )

        redis_mock = mock.MagicMock(spec=redis_module.Redis)
        redis_mock.pipeline.return_value.__enter__.return_value.get.return_value = 5
        redis_mock.pipeline.return_value.__enter__.return_value.execute.side_effect = (exceptions.WatchError, [4])
        with mock.patch.object(RedisClient, '_connection', {obj._connection_key(): redis_mock}) as _connection:
            ret = obj.decr('key', initial_value=0)

        self.assertEqual(ret, 4)

    def test_rpush(self):
        obj = RedisCache()
        self.assertEqual(obj.rpush('key', 1), True)
        self.assertEqual(obj.rpush('key', 2, 3), True)
        self.assertListEqual(obj.lrange('key', 0, -1), [b'1', b'2', b'3'])

        redis_mock = mock.MagicMock(spec=redis_module.Redis)
        redis_mock.pipeline.return_value.__enter__.return_value.execute.return_value = (4)
        with mock.patch.object(RedisClient, '_connection', {obj._connection_key(): redis_mock}) as _connection:
            obj.rpush('key', 2, timeout=1)
        _connection[obj._connection_key()].pipeline.return_value.__enter__.return_value.expire.assert_called_once_with(obj.get_hash_key('key'), 1)

    def test_hset(self):
        obj = RedisCache()
        self.assertEqual(obj.hset('key', 'field_1', 1), True)
        self.assertEqual(obj.hset('key', 'field_2', 2), True)
        self.assertEqual(obj.hset('key', 'field_3', 3), True)
        self.assertEqual(obj.hgetall('key'), {b'field_1': b'1', b'field_2': b'2', b'field_3': b'3'})
        self.assertEqual(obj.hset('key', 'field_1', 'test'), True)
        self.assertEqual(obj.hgetall('key'), {b'field_1': b'test', b'field_2': b'2', b'field_3': b'3'})

        redis_mock = mock.MagicMock(spec=redis_module.Redis)
        redis_mock.pipeline.return_value.__enter__.return_value.execute.return_value = (4)
        with mock.patch.object(RedisClient, '_connection', {obj._connection_key(): redis_mock}) as _connection:
            obj.hset('key', 'field', 2, timeout=1)
        _connection[obj._connection_key()].pipeline.return_value.__enter__.return_value.expire.assert_called_once_with(obj.get_hash_key('key'), 1)

class RedisCacheGetSetTestCase(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        RedisClient._connection = {}
        with mock.patch.object(redis_module.log, 'debug') as debug:
            cls.cache = RedisCache()
            debug.assert_called_once_with('Connecting on Redis(%s:%s).....', cls.cache.host, cls.cache.port)

        cls.key = 'redis-get-set-test-case'

    def setUp(self) -> None:
        self.cache.delete(self.key)
        self.mock = mock.MagicMock()

    def func(self, **kwargs):
        sleep(0.1)
        self.mock(**kwargs)
        return f'return: {kwargs}'

    def test_redis_get_set(self):
        obj01 = RedisCache()
        obj02 = RedisCache()
        with ThreadPoolExecutor() as executor:
            task01 = executor.submit(obj01.get_set, key=self.key, func=self.func, keyword='value01')
            sleep(0.05) # This is just to guarantee that value1 will always be called first
            task02 = executor.submit(obj02.get_set, key=self.key, func=self.func, keyword='value02')
            self.assertEqual(task01.result(), task02.result())
        self.mock.assert_called_once_with(keyword='value01')

    @mock.patch.object(redis_module.log, 'error')
    def test_get_set_func_error_both_exit(self, error: mock.MagicMock):
        """
        This test is to check if the error is raised when multiple threads
        are trying to access the same key and the function raises an error.
        The first thread will raise the error and the second thread will
        return None.
        """
        self.mock.side_effect = AttributeError
        obj01 = RedisCache()
        obj02 = RedisCache()
        with ThreadPoolExecutor() as executor:
            task01 = executor.submit(obj01.get_set, key=self.key, func=self.func, keyword='value01')
            sleep(0.05) # This is just to guarantee that value1 will always be called first
            task02 = executor.submit(obj02.get_set, key=self.key, func=self.func, keyword='value02')
            self.assertIsNone(task01.result())
            self.assertIsNone(task02.result())

        self.mock.assert_called_once_with(keyword='value01')
        error.assert_called_once()

    @mock.patch.object(redis_module.log, 'error')
    def test_get_set_func_error(self, error: mock.MagicMock):
        self.mock.side_effect = AttributeError
        result = RedisCache().get_set(key=self.key, func=self.func)
        self.assertIsNone(result)
        self.mock.assert_called_once_with()
        error.assert_called_once()


@mock.patch.object(redis_module, 'log', mock.MagicMock())
class RedisCacheCompressedTestCase(TestCase):

    def setUp(self) -> None:
        RedisClient._connection = {}

    def test_get_hash_key(self):
        obj = RedisCacheCompressed()
        hash_key = '2c70e12b7a0646f92279f427c7b38e7334d8e5389cff167a1dc30e73f826b683'
        self.assertEqual(obj.get_hash_key('key'), hash_key)
        # Must always be the same value
        self.assertEqual(obj.get_hash_key('key'), hash_key)

    def test_get_set_delete(self):
        obj = RedisCacheCompressed()
        self.assertIsNone(obj.get('key'))
        lst = [1, 'test', 1.1, {'key': 'value'}]
        obj.set('key', lst)
        self.assertListEqual(obj.get('key'), lst)
        dct = {'lst': [1, 't'], 'dct': {'key': 'value'}, 'key': 1}
        obj.set('key', dct)
        self.assertDictEqual(obj.get('key'), dct)
        obj.delete('key')
        self.assertIsNone(obj.get('key'))


@mock.patch.object(redis_module, 'log', mock.MagicMock())
class RedisListTestCase(TestCase):

    def setUp(self) -> None:
        RedisClient._connection = {}
        self.name = 'redis-list-test-case'

    def test_push_pop(self):
        lst = RedisList(name=self.name)
        lst.clear() # Clear all elements
        self.assertRaises(RedisEmptyListError, lst.pop)
        obj = {'lst': [1, 1.1, 'test', True]}
        lst.push(obj)
        self.assertEqual(lst.pop(), obj)

    def test_bpop_timeout(self):
        lst = RedisList(name=self.name)
        lst.clear() # Clear all elements
        self.assertRaises(RedisEmptyListError, lst.bpop, 0.5)

    def test_bpop(self):
        lst = RedisList(name=self.name)
        lst.clear() # Clear all elements
        obj = {'lst': [1, 1.1, 'test', True]}
        lst.push(obj)
        self.assertTupleEqual(lst.bpop(0.5), (self.name, obj))

    def test_clear(self):
        lst = RedisList(name=self.name)
        lst.clear() # Clear all elements
        self.assertRaises(RedisEmptyListError, lst.pop)
        obj = {'lst': [1, 1.1, 'test', True]}
        lst.push(obj)
        lst.clear()
        self.assertRaises(RedisEmptyListError, lst.pop)


@mock.patch.object(redis_module, 'log', mock.MagicMock())
@mock.patch.object(RedisChannel, '_channel', spec=client.PubSub)
class RedisChannelTestCase(TestCase):

    def setUp(self) -> None:
        self.name = 'tests'
        RedisClient._connection = {}

    def test_send(self, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        redis_mock = mock.MagicMock(spec=redis_module.Redis)
        with mock.patch.object(RedisClient, '_connection', {channel._connection_key(): redis_mock}) as _connection:
            channel.send('Hi')
        _connection[channel._connection_key()].publish.assert_called_once_with(channel.name, 'Hi')
        _channel.assert_not_called()

    def test_channel(self, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        redis_mock = mock.MagicMock(spec=redis_module.Redis)
        redis_mock.pubsub.return_value = _channel
        with mock.patch.object(RedisClient, '_connection', {channel._connection_key(): redis_mock}) as _connection:
            channel._channel = None
            channel.channel # Call the property - pylint: disable=pointless-statement
        _connection[channel._connection_key()].pubsub.assert_called_once_with()
        _channel.subscribe.assert_called_once_with(channel.name)

    def test_parse_message(self, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        message = {'channel': self.name.encode('utf-8'), 'data': b'data'}
        self.assertTupleEqual(channel.parse_message(message), (self.name, 'data'))
        _channel.assert_not_called()

    def test_consume_process_message(self, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        message = {'channel': self.name.encode('utf-8'), 'data': b'data'}
        _channel.listen.return_value = [message]
        channel.process_message = mock.MagicMock()
        channel.consume()
        channel.process_message.assert_called_once_with('data')

    def test_process(self, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        self.assertIsNone(channel.process_message('data'))
        _channel.assert_not_called()

    def test_consume_callback(self, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        message = {'channel': self.name.encode('utf-8'), 'data': b'data'}
        _channel.listen.return_value = [message]
        callback = mock.MagicMock()
        channel.consume(callback)
        callback.assert_called_once_with('data')

    def test_consume_break(self, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        message = {'channel': self.name.encode('utf-8'), 'data': b'data'}
        exit_message = {'channel': self.name.encode('utf-8'), 'data': RedisChannel.exit_message}
        _channel.listen.return_value = [exit_message, message]
        callback = mock.MagicMock()
        channel.consume(callback)
        callback.assert_not_called()


@mock.patch.object(redis_module, 'log', mock.MagicMock())
class RedisLockTestCase(TestCase):

    def setUp(self) -> None:
        self.name = 'tests'
        RedisClient._connection = {}

    def tearDown(self) -> None:
        RedisLock(name=self.name).release(force=True)

    def test_acquire(self):
        lock01 = RedisLock(name=self.name)
        lock02 = RedisLock(name=self.name)
        self.assertTrue(lock01.acquire(blocking=False))
        self.assertFalse(lock02.acquire(blocking=False))
        self.assertIsNone(lock01.release())
        self.assertIsNone(lock01.token)

        self.assertTrue(lock01.acquire(token='test', blocking=False))
        self.assertEqual(lock01.token, 'test')

    @mock.patch('everysk.core.redis.RedisLock._get_lock', spec=redis_module.Redis.lock)
    def test_acquire_blocking(self, lock: mock.MagicMock):
        lock01 = RedisLock(name=self.name)
        lock01.acquire(blocking_timeout=1.0)
        lock.return_value.acquire.assert_called_once_with(token=lock01._encode_token(), blocking=None, blocking_timeout=1.0)

    def test_timeout(self):
        lock01 = RedisLock(name=self.name, timeout=1)
        lock02 = RedisLock(name=self.name, timeout=1)
        self.assertTrue(lock01.acquire(blocking=False))
        self.assertFalse(lock02.acquire(blocking=False))
        sleep(1.5) # With 1 second the test breaks sometimes
        self.assertTrue(lock02.acquire(blocking=False))

    def test_owned(self):
        lock01 = RedisLock(name=self.name)
        lock02 = RedisLock(name=self.name)
        lock01.acquire(blocking=False)
        lock02.acquire(blocking=False)
        self.assertTrue(lock01.owned())
        self.assertFalse(lock02.owned())

    def test_release(self):
        lock01 = RedisLock(name=self.name)
        lock02 = RedisLock(name=self.name)
        lock01.acquire(blocking=False)
        lock02.acquire(blocking=False)
        self.assertRaisesRegex(
            exceptions.LockError,
            'Cannot release an unlocked lock',
            lock02.release
        )
        self.assertFalse(lock02.acquire(blocking=False))
        self.assertIsNone(lock01.release())
        self.assertTrue(lock02.acquire(blocking=False))

    def test_release_force(self):
        lock01 = RedisLock(name=self.name)
        lock02 = RedisLock(name=self.name)
        lock01.acquire(blocking=False)
        lock02.acquire(blocking=False)
        self.assertIsNone(lock02.release(force=True))
        self.assertTrue(lock02.acquire(blocking=False))
        self.assertRaisesRegex(
            exceptions.LockError,
            'Cannot release a lock that\'s no longer owned',
            lock01.release
        )
        self.assertFalse(lock01.acquire(blocking=False))

    def test_encode_token(self):
        lock = RedisLock(name=self.name)
        self.assertEqual(lock._encode_token('token'), b'redis-lock:3c469e9d6c5875d37a43f353d4f88e61fcf812c66eee3457465a40b0da4153e0')
        self.assertRaisesRegex(
            exceptions.LockError,
            'Cannot encode an empty token',
            lock._encode_token,
        )

    def test_do_release(self):
        lock01 = RedisLock(name=self.name)
        lock02 = RedisLock(name=self.name)
        self.assertTrue(lock01.acquire(token='test', blocking=False))
        self.assertIsNone(lock02.do_release(expected_token='test'))
        self.assertTrue(lock02.acquire(token='test', blocking=False))
        self.assertRaisesRegex(
            exceptions.LockError,
            'Cannot release an unlocked lock',
            lock01.do_release,
            expected_token=None
        )
        lock02.name = 'test2'
        self.assertIsNone(lock02.do_release(expected_token='token_2'))

    def test_get_lock_info(self):
        lock = RedisLock(name=self.name)
        self.assertEqual(lock.get_lock_info(), {
            'locked': False,
            'name': lock.name
        })
        self.assertTrue(lock.acquire(blocking=False))
        self.assertEqual(lock.get_lock_info(), {
            'locked': True,
            'name': lock.name
        })



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from types import UnionType
from typing import Any

from everysk.core.exceptions import FieldValueError, RequiredError


class Field:
    ## Private attributes
    _empty_values: frozenset[Any] = frozenset({None, Undefined, '', (), [], {}})
    _always_valid: frozenset[Any] = frozenset({None, Undefined, callable})

    ## Public attributes
    attr_name: str = None
    attr_type: type | UnionType = None
    choices: set[Any] = None
    default: Any = None
    max_value: Any = None
    min_value: Any = None
    readonly: bool = False
    required: bool = False
    required_lazy: bool = False
    empty_is_none: bool = False

    ## Internal methods
    def __init__(self, default: Any = None, *, readonly: bool = False, required: bool = False, **kwargs) -> None:
        if required and kwargs.get('required_lazy'):
            msg = f"{self.attr_name} -> required and required_lazy can't be both True."
            raise FieldValueError(msg)
        self.required = required

        if readonly and default in self._empty_values:
            msg = f'{self.attr_name} -> is readonly then default value is required.'
            raise RequiredError(msg)
        self.readonly = readonly

        # Set all other attributes if they exist in the class
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                msg = f'{self.attr_name} -> invalid attribute: {key}'
                raise FieldValueError(msg)

        # We store the default value already cleaned
        self.default = self.clean_value(default)

    def __get__(self, obj: object, cls: type) -> Any:
        # We will use this var to store the correct place to get the attr_name value
        dct = {}
        # If it is a dict we get it directly
        if isinstance(obj, dict):
            dct = obj
        else:
            # All python objects have the __dict__ attribute except for builtins
            dct = getattr(obj, '__dict__', {})

        # If the value is not in the dict we try to get it from the class
        # We need to use getattr here to raise AttributeError if we can't find the attribute
        value = dct.get(self.attr_name, getattr(cls, self.attr_name))

        if callable(value):
            return value()

        return value

    def __set__(self, obj: object, value: Any) -> None:

    ## Public methods
    def clean_value(self, value: Any) -> Any:
        return value

    def validate(self, value: Any) -> None:
        if self.readonly and value != self.default:
            # This is necessary to be able to at least assign the default value to the field
            msg = f"The field '{self.attr_name}' value cannot be changed."
            raise FieldValueError(msg)

        if self.required and value in self._empty_values:
            msg = f'The {self.attr_name} attribute is required.'
            raise RequiredError(msg)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
import traceback
from hashlib import sha256
from functools import wraps, _HashedSeq
from time import sleep
from uuid import uuid1
from typing import Any, Iterable

from redis import Redis, client, exceptions # pylint: disable=import-self
from redis.backoff import ExponentialBackoff # pylint: disable=import-error, no-name-in-module
from redis.lock import Lock # pylint: disable=import-error, no-name-in-module
from redis.retry import Retry # pylint: disable=import-error, no-name-in-module

from everysk.config import settings
from everysk.core.compress import compress, decompress
from everysk.core.exceptions import RedisEmptyListError
from everysk.core.fields import FloatField, StrField, BoolField
from everysk.core.log import Logger
from everysk.core.object import BaseObject
from everysk.core.serialize import dumps, loads

log = Logger(name='everysk-redis')

###############################################################################
#   Cache decorator functions Implementation
###############################################################################
def _make_key(args: tuple, kwargs: dict) -> str:
    """
    Create a key from args and kwargs to be used on cache.
    This function is a based on functools._make_key

    Args:
        args (tuple): The received args.
        kwargs (dict): The received kwargs.
    """
    #/usr/local/lib/python3.11/functools.py: 448 - def _make_key
    key = args
    if kwargs:
        for item in kwargs.items():
            key += item
    elif len(key) == 1 and type(key[0]) in {int, float, str}:
        # To be faster if we have only one item and it's a int, float or str
        # we return it directly
        return str(key[0])

    return str(_HashedSeq(key))

def cache(*, timeout: int):
    """
    Decorator to cache the result of a function in Redis.
    To disable the cache, set the timeout to None otherwise it must be > 0.

    Args:
        timeout (int): The time in seconds that the result will be stored in cache.

    Raises:
        ValueError: If timeout is not an integer or is less than 0.
    """
    if timeout is not None and (not isinstance(timeout, int) or timeout < 1):
        raise ValueError('Timeout must be an integer greater than 0.')

    # Because we have a parameter we need to create another nested function
    def decorator(func):
        info = {'hits': 0, 'misses': 0}
        redis_cache = RedisCache(prefix=func.__name__)

        # We use wraps to keep the original function name and docstring
        @wraps(func)
        def wrapper(*args, **kwargs):
            key = _make_key(args, kwargs)
            value = redis_cache.get(key)
            if value:
                info['hits'] += 1
                return loads(value, protocol='pickle')

            info['misses'] += 1
            value = func(*args, **kwargs)
            redis_cache.set(key, dumps(value, protocol='pickle'), timeout)
            return value

        # We add some extra attributes to the wrapper
        wrapper.info = info
        wrapper.clear = redis_cache.delete_prefix

        return wrapper

    return decorator

###############################################################################
#   RedisClient Class Implementation
###############################################################################
class RedisClient(BaseObject):
    ## Private attributes
    _connection: dict = {}
    _separator = ':'

    ## Public attributes
    timeout_default: int = None
    prefix: str = None
    host: str = None
    port: int = None

    ## Private methods
    def _connect(self) -> None:
        """ Create a Redis connection and stores to later use. """
        # https://redis-py.readthedocs.io/en/stable/retry.html
        retry = Retry(ExponentialBackoff(), 3)

        # https://github.com/redis/redis-py/issues/722
        # We use RedisClient._connection to create a Singleton connection
        log.debug('Connecting on Redis(%s:%s).....', self.host, self.port)

        RedisClient._connection[self._connection_key()] = Redis(
            host=self.host,
            port=self.port,
            health_check_interval=settings.REDIS_HEALTH_CHECK_INTERVAL, # seconds
            socket_keepalive=settings.REDIS_SOCKET_KEEPALIVE,
            socket_timeout=settings.REDIS_SOCKET_TIMEOUT, # seconds
            retry=retry,
            retry_on_error=[Exception]
        )

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        if self.host is None:
            self.host = settings.REDIS_HOST

        if self.port is None:
            self.port = settings.REDIS_PORT

        if RedisClient._connection.get(self._connection_key(), None) is None:
            self._connect()

    @property
    def connection(self) -> Redis:
        """
        We use this property to check if Redis is online
        then returning the working connection.
        """
        try:
            RedisClient._connection[self._connection_key()].ping()
        except Exception: # pylint: disable=broad-exception-caught
            # Create a new connection
            self._connect()

        return RedisClient._connection[self._connection_key()]

    def _connection_key(self) -> str:
        return f'{self.host}:{self.port}'

    def _build_prefix(self, prefix: str | None = None) -> str:
        """
        Build the prefix for the Redis key using the namespace and prefix.

        Args:
            prefix (str | None): The prefix to use. If None, use the class prefix.

        Returns:
            str: The constructed prefix.
        """
        if prefix is None:
            prefix = self.prefix

        parts = []
        if settings.REDIS_NAMESPACE:
            parts.append(settings.REDIS_NAMESPACE)
        if prefix:
            parts.append(prefix)

        return self._separator.join(parts)

    def _build_key(self, key: str | None = None, parts: list | None = None) -> str:
        """
        Build the key for the Redis key using the prefix, namespace and key.

        Args:
            key (str | None): The key to use. If None, use the class key.
            parts (list | None): The list of parts to use.

        Returns:
            str: The constructed key.
        """
        parts = parts or []
        prefix = self._build_prefix()
        if prefix:
            parts.append(prefix)
        if key:
            parts.append(key)

        return self._separator.join(parts)

    def _encode(self, value: Any) -> bytes:
        """
        Encodes a value before storing it in Redis.

        This method can be overridden by subclasses when specific serialization
        or transformation logic is needed before saving the value to Redis. By centralizing
        this behavior here, other methods that interact with Redis do not need to be modified.

        Args:
            value (Any): The value to be stored.

        Returns:
            bytes: The serialized or transformed value ready for storage.
        """
        return value

    def _decode(self, value: bytes) -> Any:
        """
        Decodes a value retrieved from Redis.

        This method can be overridden by subclasses when specific deserialization
        or transformation is required after retrieving the value from Redis. This provides
        a single point of customization without needing to override higher-level logic.

        Args:
            value (bytes): The raw value retrieved from Redis.

        Returns:
            Any: The original or transformed value.
        """
        return value

    def get_hash_key(self, key: bytes | str) -> str:
        """
        Convert the key to a SHA256 hash to avoid strange chars on name that can broke Redis.
        This method adds the prefix and namespaces to the key.

        Args:
            key (bytes | str): The original key to hash.

        Returns:
            str: The fully constructed and hashed Redis key.
        """
        if key is None:
            raise ValueError('Key cannot be None.')

        if isinstance(key, str):
            key = key.encode('utf-8')

        hashed_key = sha256(key).hexdigest()

        return self._build_key(hashed_key)

    def flush_all(self) -> bool:
        """ Clear all keys from Redis. """
        if self.connection.flushall():
            if settings.REDIS_SHOW_LOGS:
                log.info('Redis flushed all keys.')
            return True
        else:
            log.error('Redis flush all keys failed.')
            return False

###############################################################################
#   RedisCache Class Implementation
###############################################################################
class RedisCache(RedisClient):
    """ Redis cache client """

    def get_set(self, key: str, func: callable, timeout: int = None, **kwargs) -> Any:
        """
        This method execute a get on Redis server, if this returns None then we execute the set method.
        For the set method we use the func(**kwargs).
        To avoid race conditions we use a RedisLock to run only one set method.

        Args:
            key (str): The key that will be used to cache the result from func.
            func (callable): The function that generates the desired cached result.
            timeout (int, optional): The timeout that this key will be keep on cache. Defaults to None.
            *kwargs (dict, optional): Extra params that will be send to the func.

        Returns:
            Any: The result from cache get or the result from func.
        """
        result = self.get(key)
        if result is None:
            # We create a lock on Redis
            lock = RedisLock(name=f'redis-get-set-lock-{key}')
            if lock.acquire(blocking=False):
                # If we can acquire the lock then we don't have race conditions
                # and proceed normally
                try:
                    result = func(**kwargs)
                    self.set(key=key, value=result, timeout=timeout)
                except Exception: # pylint: disable=broad-exception-caught
                    # We generate a log for this exception
                    log.error('Redis get_set method: %s', traceback.format_exc())

                # Then we release the the lock
                lock.release()

            else:
                # If we can't acquire the lock that means we have race condition
                # in this case we need to wait for the key be set or if some error
                # occur the lock will be released.
                result = self.get(key)
                while result is None:
                    result = self.get(key)
                    if result is None and lock.acquire(blocking=False):
                        # Then we release the the lock and exit
                        lock.release()
                        break

                    # We wait 0.5 second until next try
                    sleep(0.5)

        if isinstance(result, bytes):
            result = result.decode('utf-8')

        return result

    def get(self, key: bytes) -> Any:
        """
        Get the value from Redis using the key.

        Args:
            key (bytes): The key to retrieve the value.

        Returns:
            bytes: The value associated with the key.
        """
        key = self.get_hash_key(key)
        value = self.connection.get(key)
        return self._decode(value)

    def get_multi(self, keys: list[str]) -> dict:
        """
        Get multiple values from Redis using a list of keys.

        Args:
            keys (list[str]): The list of keys to retrieve values for.

        Returns:
            dict: A dictionary containing the keys and their corresponding values.
        """
        keys_ = [self.get_hash_key(key) for key in keys]
        data_list = self.connection.mget(keys_)
        return {key: self._decode(value) for key, value in zip(keys, data_list)}

    def set(self, key: bytes, value: Any, timeout: int = None) -> bool:
        """
        Set key/value on connection for timeout in seconds,
        if timeout is None the key/value will be keep forever.
        Value must be one of these: bytes, str, int or float.

        Args:
            key (bytes): The key to set.
            value (Any): The value to set.
            timeout (int, optional): The timeout in seconds. Defaults to None.

        Raises:
            ValueError: If the value is not one of the allowed types.

        Returns:
            bool: True if the operation was successful, False otherwise.
        """
        key = self.get_hash_key(key)
        value = self._encode(value)
        ret = self.connection.set(name=key, value=value, ex=timeout or self.timeout_default)
        return bool(ret)

    def set_multi(self, data_dict: dict, timeout: int = None) -> None:
        timeout = timeout or self.timeout_default

        pipe = self.connection.pipeline()
        for key, data in data_dict.items():
            pipe.set(self.get_hash_key(key), self._encode(data), ex=timeout)
        ret = pipe.execute()

        if False in ret:
            log.error('Error RedisCache set_multi', extra={
                'labels': {
                    'REDIS_NAMESPACE': settings.REDIS_NAMESPACE,
                    'data': data_dict,
                    'time': timeout
                }
            })

        return None

    def delete(self, key: bytes | str | Iterable) -> bool:
        """
        Delete one or more keys from Redis.

        Args:
            key (bytes | str | Iterable): The key or keys to be deleted.

        Returns:
            bool: True if the operation was successful, False otherwise.
        """
        if not isinstance(key, (list, set, tuple)):
            keys = [self.get_hash_key(key)]
        else:
            keys = [self.get_hash_key(k) for k in key]

        if not keys:
            return False

        ret = self.connection.delete(*keys) == len(keys)
        return bool(ret)

    def delete_multi(self, keys: Iterable) -> bool:
        """
        Delete multiple keys from Redis.

        Args:
            keys (Iterable): The keys to be deleted.

        Returns:
            bool: True if the operation was successful, False otherwise.
        """
        return self.delete(keys)

    def delete_prefix(self, prefix: str = None) -> None:
        """
        Delete all keys with the given prefix.
        If prefix is None, we use the prefix from the class.
        To search for all keys with this prefix we need to add a ':*' at the end.
        We use the scan method to search for all keys with this prefix and pipe to delete them.

        Args:
            prefix (str | None): The prefix to use. If None, use the class prefix.
        """
        # If prefix is None, we use the prefix from the class
        # To search for all keys with this prefix we need to add a ':*' at the end
        prefix = self._build_prefix(prefix)
        if '*' not in prefix:
            prefix = f'{prefix}{self._separator}*'

        # We use the scan method to search for all keys with this prefix and pipe to delete them
        pipe = self.connection.pipeline()
        cursor = None
        while cursor != 0:
            cursor, keys = self.connection.scan(cursor=cursor or 0, match=prefix, count=1000)
            if keys:
                pipe.delete(*keys)

        pipe.execute()

    def incr(self, key: str, delta: int = 1, initial_value: Any = None, timeout: int | None = None) -> int:
        """
        Increment the value of a key in Redis. If the key does not exist, it will be created with the initial value.

        Args:
            key (str): The key to increment.
            delta (int, optional): The amount to increment. Defaults to 1.
            initial_value (Any, optional): The initial value to set if the key does not exist. Defaults to None.
            timeout (int | None, optional): The expiration time for the key in seconds. Defaults to None.

        Returns:
            int: The new value of the key after incrementing.
        """
        if initial_value is None:
            raise ValueError('Initial value must be set.')

        key_ = self.get_hash_key(key)

        pipe = self.connection.pipeline()
        pipe.multi()
        pipe.set(key_, initial_value, nx=True, ex=timeout or self.timeout_default)
        pipe.incr(key_, delta)
        ret = pipe.execute()

        if ret[1] is None:
            log.error('Error RedisCache incr', extra={
                'labels': {
                    'REDIS_NAMESPACE': settings.REDIS_NAMESPACE,
                    'key': key,
                    'delta': delta,
                    'initial_value': initial_value
                }
            })

        return ret[1]

    def decr(self, key: str, delta: int = 1, initial_value: Any = None, timeout: int = None) -> int | None:
        """
        Decrement the value of a key in Redis. If the key does not exist, it will be created with the initial value.

        Args:
            key (str): The key to decrement.
            delta (int, optional): The amount to decrement. Defaults to 1.
            initial_value (Any, optional): The initial value to set if the key does not exist. Defaults to None.
            timeout (int | None, optional): The expiration time for the key in seconds. Defaults to None.

        Returns:
            int | None: The new value of the key after decrementing, or None if the operation failed.
        """
        if initial_value is None:
            raise ValueError('Initial value must be set.')

        key_ = self.get_hash_key(key)
        with self.connection.pipeline(True) as pipe:
            while True:
                try:
                    pipe.watch(key_)
                    current = self._decode(pipe.get(key_))

                    if isinstance(current, bytes):
                        current = int(current.decode())

                    if current is None:
                        current = 0

                    pipe.multi()
                    if current >= delta:
                        pipe.decr(key_, delta)
                    else:
                        value = initial_value - delta if current == 0 else current - delta
                        value = 0 if value <= 0 else value
                        pipe.set(key_, value, ex=timeout or self.timeout_default)
                    exec_value = pipe.execute()
                    break
                except exceptions.WatchError:
                    continue

        return value if exec_value[0] is True else exec_value[0]

    def blpop(self, keys: list, timeout: int = 0) -> tuple:
        """
        Pop the first item from the list, blocking until a item exists
        or timeout was reached.
        If timeout is 0, then block indefinitely.

        Args:
            keys (list): The list of keys to pop from.
            timeout (int, optional): The timeout in seconds. Defaults to 0.

        Returns:
            tuple: (name, value)
        """
        if not isinstance(keys, list):
            keys = [keys]
        keys_mapping = {self.get_hash_key(key): key for key in keys}
        values = self.connection.blpop(list(keys_mapping.keys()), timeout=timeout)
        if values is None:
            return None

        return (keys_mapping[values[0].decode()], self._decode(values[1]))

    def lpop(self, name: str, count: int | None = None) -> Any:
        """
        Pop the first item from the list.

        Args:
            name (str): The name of the list.
            count (int | None, optional): The number of items to pop. Defaults to None.

        Returns:
            Any: The popped value.
        """
        value = self.connection.lpop(self.get_hash_key(name), count=count)
        return self._decode(value)

    def rpush(self, name: str, *values, timeout: int | None = None) -> bool:
        """
        Push one or more values to the end of the list.
        If the list does not exist, it will be created.
        If the timeout is None, the list will be kept forever.
        If the timeout is > 0, the list will be kept for the specified time.

        Args:
            name (str): The name of the list.
            *values: The values to push to the list.
            timeout (int | None, optional): The expiration time for the list in seconds. Defaults to None.

        Returns:
            bool: True if the operation was successful, False otherwise.
        """
        key = self.get_hash_key(name)
        timeout = timeout or self.timeout_default

        datas = []
        for value in values:
            datas.append(self._encode(value))

        with self.connection.pipeline() as pipe:
            pipe.rpush(key, *datas)
            if timeout:
                pipe.expire(key, timeout)
            ret = pipe.execute()

        return bool(ret)

    def lrange(self, name: str, start: int, end: int) -> list:
        """
        Get a range of values from the list.

        Args:
            name (str): The name of the list.
            start (int): The starting index.
            end (int): The ending index.

        Returns:
            list: The list of values in the specified range.
        """
        key = self.get_hash_key(name)
        ret = self.connection.lrange(key, start, end)

        out = []
        for data in ret:
            out.append(self._decode(data))

        return out

    def hset(self, name: str, field: str, value: Any, timeout: int | None = None) -> bool:
        """
        Set a field in a hash stored at key.
        If the key does not exist, a new key will be created.
        If the field already exists, it will be overwritten.
        If the timeout is None, the hash will be kept forever.
        If the timeout is > 0, the hash will be kept for the specified time.

        Args:
            name (str): The name of the hash.
            field (str): The field to set.
            value (Any): The value to set.
            timeout (int | None, optional): The expiration time for the hash in seconds. Defaults to None.

        Returns:
            bool: True if the operation was successful, False otherwise.
        """
        key = self.get_hash_key(name)
        timeout = timeout or self.timeout_default

        value = self._encode(value)
        with self.connection.pipeline() as pipe:
            pipe.hset(key, field, value)
            if timeout:
                pipe.expire(key, timeout or self.timeout_default)
            ret = pipe.execute()

        return bool(ret)

    def hgetall(self, name: str) -> dict:
        """
        Get all fields and values in a hash stored at key.
        If the key does not exist, an empty dictionary will be returned.
        If the key is not a hash, an error will be raised.

        Args:
            name (str): The name of the hash.

        Returns:
            dict: A dictionary containing all fields and values in the hash.
        """

        key = self.get_hash_key(name)
        ret = self.connection.hgetall(key)

        out = {}
        for field, data in ret.items():
            out[field] = self._decode(data)

        return out

###############################################################################
#   RedisCacheCompressed Class Implementation
###############################################################################
class RedisCacheCompressed(RedisCache):
    """
    Store data on Redis server using pickle and zlib
    Use this if you need to store objects ons Redis.
    """

    def _encode(self, value: Any) -> bytes:
        """
        Encode the value to bytes using pickle and zlib.

        Args:
            value (Any): The value to encode.

        Returns:
            bytes: The encoded value.
        """
        return compress(value, serialize='pickle') if value else None

    def _decode(self, value: bytes) -> Any:
        """
        Decode the value from bytes using pickle and zlib.

        Args:
            value (bytes): The value to decode.

        Returns:
            Any: The decoded value.
        """
        return decompress(value, serialize='pickle') if value else None

###############################################################################
#   RedisList Class Implementation
###############################################################################
class RedisList(RedisCacheCompressed):
    """
    First in, first out Redis list implementation.
    -> https://redis.io/docs/data-types/lists/
    -> https://koalatea.io/python-redis-lists/

    """
    name = StrField(required=True)

    def bpop(self, timeout: int = 0) -> tuple:
        """
        Pop the first item from the list, blocking until a item exists
        or timeout was reached.
        If timeout is 0, then block indefinitely.

        Returns:
            tuple: (list name, value)
        """
        value = super().blpop(self.name, timeout=timeout)
        if value is None:
            raise RedisEmptyListError(f"The RedisList(name='{self.name}') is empty.")
        return value

    def pop(self) -> Any:
        """
        Pop the first item from the list.

        Raises:
            RedisEmptyListError: If the return is None/empty.
        """
        value = super().lpop(self.name)
        if value is None:
            raise RedisEmptyListError(f"The RedisList(name='{self.name}') is empty.")

        return value

    def push(self, value: Any) -> None:
        """
        Puts value on the last position of the list.

        Args:
            value (Any): the value to be inserted into the last position
        """
        super().rpush(self.name, *[value])

    def clear(self) -> None:
        """
        Clear all keys
        """
        super().delete(self.name)

###############################################################################
#   RedisChannel Class Implementation
###############################################################################
class RedisChannel(RedisClient):
    """
    Base class to work with channels on Redis.
    https://blog.devgenius.io/how-to-use-redis-pub-sub-in-your-python-application-b6d5e11fc8de
    """
    _channel: client.PubSub = None
    exit_message = StrField(default='exit', readonly=True)
    name = StrField(required=True)

    def send(self, message: dict) -> None:
        self.connection.publish(self.name, message)

    @property
    def channel(self) -> client.PubSub:
        """ Create a connection with name """
        if self._channel is None:
            self._channel = self.connection.pubsub()
            self._channel.subscribe(self.name)

        return self._channel

    def parse_message(self, message: dict) -> tuple:
        """
        Convert message data from bytes to str

        Args:
            message (dict): The message dictionary to be parsed.

        Returns:
            tuple: Containing the channel name and its corresponding data.
        """
        # message format
        # {'type': None, 'pattern': None, 'channel': None, 'data': None}
        channel_name = None
        data = None
        if message:
            channel_name = message.get('channel', None) or None
            data = message.get('data', '') or ''
            if isinstance(channel_name, bytes):
                channel_name = channel_name.decode()
            if isinstance(data, bytes):
                data = data.decode()

        return (channel_name, data)

    def consume(self, callback: callable = None) -> None:
        """ Loop for consume message from channel when they arrive. """
        for message in self.channel.listen():
            channel_name, data = self.parse_message(message)
            # Only care if the message is sent to this channel
            if channel_name == self.name:
                # Stop iteration on exit_message
                if data == self.exit_message:
                    break

                # We can use a function for callback or self.process_message
                if callback:
                    callback(data)
                else:
                    self.process_message(data)

    def process_message(self, message: str) -> None:
        """ Use it on child classes to manipulate the received message. """
        pass

###############################################################################
#   RedisLock Class Implementation
###############################################################################
class RedisLock(RedisClient):
    """
    Class used to create a lock on Redis
    https://rohansaraf.medium.com/distributed-locking-with-redis-ecb0773e7695
    https://redis-py.readthedocs.io/en/latest/lock.html
    """
    ## Public attributes
    prefix = StrField(default='redis-lock', readonly=True)
    token = StrField()
    name = StrField(required=True)
    timeout = FloatField(default=None) # timeout indicates a maximum life for the lock in seconds.
    blocking = BoolField(default=True) # If True, the lock will block until it can be acquired.

    def _get_lock(self) -> Lock:
        """
        Create a lock object with the name and timeout.
        If the token is set, it will be encoded.
        The token is used to identify the lock owner.
        The lock is created with the name and timeout.
        The timeout is the maximum life for the lock in seconds.
        If the timeout is None, the lock will be kept forever.

        Returns:
            Lock: The lock object.
        """
        lock = self.connection.lock(
            name=self._get_name(),
            timeout=self.timeout,
            blocking=self.blocking
        )

        if self.token:
            lock.local.token = self._encode_token()

        return lock

    def _encode_token(self, token: str | None = None) -> bytes:
        """
        Encode the token to bytes using the Redis encoder.
        This is used to identify the lock owner.

        Args:
            token (str | None): The token to encode. If None, use the class token.

        Raises:
            LockError: If the token is None or empty.

        Returns:
            bytes: The encoded token.
        """
        token = token or self.token
        if not token:
            raise exceptions.LockError('Cannot encode an empty token')

        redis_encoder = self.connection.get_encoder()
        return redis_encoder.encode(self._get_token(token))

    def _get_name(self) -> str:
        """
        Convert self.name to a SHA256 hash, this avoid strange chars on name that can broke Redis.
        """
        return self.get_hash_key(self.name)

    def _get_token(self, token: str | bytes | None = None) -> str:
        """
        Convert self.token to a SHA256 hash, this avoid strange chars on token that can broke Redis.
        """
        return self.get_hash_key(token or self.token)

    def acquire(self, token: str | None = None, blocking: bool = None, blocking_timeout: float = None) -> bool:
        """
        Try to acquire a lock with self.name, if lock is already acquired returns False.
        If blocking is False, always return immediately, if blocking is True it will waiting until block can be acquired.
        blocking_timeout specifies the maximum number of seconds to wait trying to acquire the lock.
        If token is None, a new token will be generated.
        If token is not None, the lock will be acquired with this token.
        This token is used to identify the lock owner.

        Args:
            token (str | None): The token to use for the lock. If None, a new token will be generated.
            blocking (bool | None): If True, wait until the lock is acquired. Defaults to None.
            blocking_timeout (float | None): The maximum number of seconds to wait for the lock. Defaults to None.

        Returns:
            bool: True if the lock was acquired, False otherwise.
        """
        if not token and not self.token:
            self.token = uuid1().hex
        elif token:
            self.token = token

        redis_lock: Lock = self._get_lock()

        if redis_lock.acquire(token=self._encode_token(), blocking=blocking, blocking_timeout=blocking_timeout):
            return True

        # If we can't acquire the lock, we need to check if the lock is owned by this instance
        self.token = None
        return False

    def owned(self) -> bool:
        """
        Returns True if this key is locked by this lock, otherwise False.
        This method is used to check if the lock is owned by this instance.

        Returns:
            bool: True if the lock is owned by this instance, False otherwise.
        """
        if not self.token:
            return False
        redis_lock: Lock = self._get_lock()
        redis_lock.local.token = self._encode_token()
        return redis_lock.owned()

    def release(self, force: bool = False) -> None:
        """
        Release the lock if it is owned by this instance.
        If the lock is not owned, it will raise a LockError.
        This method is used to release the lock and set the token to None.

        Args:
            force (bool): If True, force the release of the lock even if it is not owned.
                This will delete the key from Redis. Defaults to False.
        """
        if force:
            self.connection.delete(self._get_name())
            return None

        if not self.token:
            raise exceptions.LockError('Cannot release an unlocked lock')

        redis_lock: Lock = self._get_lock()
        redis_lock.release()
        self.token = None

        return None

    def do_release(self, expected_token: str) -> None:
        """
        Force release lock by an token.

        Args:
            token (str): lock key token.
        """
        if not expected_token:
            raise exceptions.LockError('Cannot release an unlocked lock')

        redis_lock = self._get_lock()
        if redis_lock.locked():
            redis_lock.do_release(expected_token=self._encode_token(expected_token))

    def get_lock_info(self) -> dict:
        """
        Get information about the lock status.
        This includes whether the lock is currently held and the name of the lock.

        Returns:
            dict: A dictionary containing the lock status and name.
        """
        redis_lock = self._get_lock()
        return {
            'locked': redis_lock.locked(),
            'name': self.name
        }



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.http import httpx
from everysk.core.compress import compress
from everysk.core.exceptions import InvalidArgumentError
from everysk.core.unittests import TestCase, mock
from everysk.sdk.base import HttpSDKPOSTConnection
from everysk.sdk.engines.lock import UserLock

###############################################################################
#   Get Method Test Case Implementation
###############################################################################
class UserLockTestCase(TestCase):

    def setUp(self):
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()

    def test_initialization(self):
        user_lock = UserLock(name='test_lock')
        self.assertIsInstance(user_lock, UserLock)
        self.assertEqual(user_lock.name, 'test_lock')
        self.assertEqual(user_lock.timeout, settings.USER_CACHE_LOCK_EXPIRATION_TIME)
        self.assertTrue(user_lock.blocking)
        self.assertIsNone(user_lock.token)

        user_lock = UserLock(name='test_lock')
        user_lock.name = 'test_lock_1'
        user_lock.timeout = 20.0
        user_lock.blocking = False
        user_lock.token = 'test_token'
        self.assertEqual(user_lock.name, 'test_lock_1')
        self.assertEqual(user_lock.timeout, 20.0)
        self.assertFalse(user_lock.blocking)
        self.assertEqual(user_lock.token, 'test_token')

        user_lock = UserLock(name='test_lock', timeout=15.0, blocking=False, token='test_token')
        self.assertEqual(user_lock.name, 'test_lock')
        self.assertEqual(user_lock.timeout, 15.0)
        self.assertFalse(user_lock.blocking)
        self.assertEqual(user_lock.token, 'test_token')

    def test_acquire(self):
        test_name = 'port_xyz'
        user_cache_lock: UserLock = UserLock(name=test_name)
        with mock.patch('httpx.Client.post') as mock_get_response:
            mock_get_response.return_value.content = 'true'
            mock_get_response.return_value.status_code = 200
            ret = user_cache_lock.acquire()

        self.assertTrue(ret)
        mock_get_response.assert_called_once_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            content=compress({'class_name': 'UserLock', 'method_name': 'acquire', 'self_obj': user_cache_lock.to_dict(add_class_path=True), 'params': {'token': None, 'blocking': None, 'blocking_timeout': None}}, protocol='gzip', serialize='json')
        )

        test_name = 'port_xyz'
        test_token = 'test_token'
        user_cache_lock: UserLock = UserLock(name=test_name)
        with mock.patch('httpx.Client.post') as mock_get_response:

            mock_get_response.return_value.content = 'true'
            mock_get_response.return_value.status_code = 200
            ret = user_cache_lock.acquire(token=test_token)

        self.assertTrue(ret)
        mock_get_response.assert_called_once_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            content=compress({'class_name': 'UserLock', 'method_name': 'acquire', 'self_obj': user_cache_lock.to_dict(add_class_path=True), 'params': {'token':test_token, 'blocking': None, 'blocking_timeout': None}}, protocol='gzip', serialize='json')
        )

    def test_release(self):
        test_name = 'port_xyz'

        user_cache_lock: UserLock = UserLock(name=test_name)
        self.assertRaisesRegex(
            InvalidArgumentError,
            'Cannot release an unlocked lock',
            user_cache_lock.release
        )

        user_cache_lock.token = 'test_token'
        with mock.patch('httpx.Client.post') as mock_get_response:
            mock_get_response.return_value.content = 'null'
            mock_get_response.return_value.status_code = 200
            ret = user_cache_lock.release()

        self.assertIsNone(ret)
        mock_get_response.assert_called_once_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            content=compress({'class_name': 'UserLock', 'method_name': 'release', 'self_obj': user_cache_lock.to_dict(add_class_path=True), 'params': {'force': False}}, protocol='gzip', serialize='json')
        )

    def test_do_release(self):
        test_name = 'port_xyz'
        expected_token = 'test_token'
        user_cache_lock: UserLock = UserLock(name=test_name)

        self.assertRaisesRegex(
            InvalidArgumentError,
            'Cannot release an unlocked lock',
            user_cache_lock.do_release,
            expected_token=None,
        )

        with mock.patch('httpx.Client.post') as mock_get_response:
            mock_get_response.return_value.content = 'null'
            mock_get_response.return_value.status_code = 200
            ret = user_cache_lock.do_release(expected_token=expected_token)

        self.assertIsNone(ret)
        mock_get_response.assert_called_once_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            content=compress({'class_name': 'UserLock', 'method_name': 'do_release', 'self_obj': user_cache_lock.to_dict(add_class_path=True), 'params': {'expected_token': expected_token}}, protocol='gzip', serialize='json')
        )

    def test_get_lock_info(self):
        test_name = 'port_xyz'
        user_cache_lock: UserLock = UserLock(name=test_name)
        with mock.patch('httpx.Client.post') as mock_get_response:
            mock_get_response.return_value.content = '{"locked": true, "name": "port_xyz"}'
            mock_get_response.return_value.status_code = 200
            ret = user_cache_lock.get_lock_info()

        self.assertIsInstance(ret, dict)
        self.assertDictEqual(ret, {'locked': True, 'name': test_name})
        mock_get_response.assert_called_once_with(
            url=self.api_url,
            headers=self.headers,
            timeout=httpx.Timeout(timeout=30, read=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT),
            content=compress({'class_name': 'UserLock', 'method_name': 'get_lock_info', 'self_obj': user_cache_lock.to_dict(add_class_path=True), 'params': {}}, protocol='gzip', serialize='json')
        )



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from uuid import uuid1

from everysk.config import settings
from everysk.core.fields import StrField, FloatField, BoolField
from everysk.core.exceptions import InvalidArgumentError
from everysk.sdk.base import BaseSDK

###############################################################################
#   UserLock Class Implementation
###############################################################################
class UserLock(BaseSDK):
    """
    UserLock class for acquiring and releasing locks.
    This class provides methods to acquire a lock for a given token, release the lock,
    and get information about the lock status.
    It is used to manage locks in a distributed system to
    prevent multiple processes from accessing the same resource simultaneously.

    Attributes:
        token (str): The unique identifier for the lock.
        name (str): The name of the lock.
        timeout (int): The expiration time for the lock in seconds.

    Example:
        >>> user_lock = UserLock(name='my_lock')
        >>> user_lock.acquire()
        UserLock(token='generated_token', name='my_lock', timeout=30)
        >>> user_lock.get_lock_info()
        {'locked': True, 'name': 'my_lock'}
        >>> user_lock.release()
        None
    """
    token = StrField(default=None)
    name = StrField(default=None, required=True)
    timeout = FloatField(default=settings.USER_CACHE_LOCK_EXPIRATION_TIME, min_size=settings.USER_CACHE_LOCK_MIN_EXPIRATION_TIME, max_size=settings.USER_CACHE_LOCK_MAX_EXPIRATION_TIME)
    blocking = BoolField(default=True) # If True, the lock will block until it can be acquired.

    def acquire(self, token: str = None, blocking: bool = None, blocking_timeout: float = None) -> bool:
        """
        Acquire a lock for a given token.

        Args:
            token (str | None): The token to use for the lock. If None, a new token will be generated.
            blocking (bool | None): If True, wait until the lock is acquired. Defaults to None.
            blocking_timeout (float | None): The maximum number of seconds to wait for the lock. Defaults to None.

        Returns:
            bool: True if the lock was acquired successfully, False otherwise.

        Example:
            >>> user_lock = UserLock(name='my_lock')
            >>> user_lock.acquire()
            True
        """
        if not token and not self.token:
            self.token = uuid1().hex
        elif token:
            self.token = token

        return self.get_response(self_obj=self, params={'token': token, 'blocking': blocking, 'blocking_timeout': blocking_timeout})


    def release(self, force: bool = False) -> None:
        """
        Release the lock for a given identifier.

        Args:
            force (bool): If True, force the release of the lock even if it is not owned.
                This will delete the key from Redis. Defaults to False.

        Example:
            >>> UserLock.release('my_lock')
            None
        """
        if not self.token:
            raise InvalidArgumentError("Cannot release an unlocked lock")

        return self.get_response(self_obj=self, params={'force': force})

    def do_release(self, expected_token: str) -> None:
        """
        Release the lock for a given token.
        This method is used to release a lock that was previously acquired.

        Args:
            expected_token (str): The unique token for the lock to release.

        Example:
            >>> user_lock = UserLock(name='my_lock')
            >>> user_lock.do_release(token='token')
            None
        """
        if not expected_token:
            raise InvalidArgumentError("Cannot release an unlocked lock")

        return self.get_response(self_obj=self, params={'expected_token': expected_token})

    def get_lock_info(self) -> dict:
        """
        Get information about the lock status.
        This method returns a dictionary containing information about the lock status.

        Returns:
            dict: A dictionary containing the lock status and other information.

        Example:
            >>> user_lock = UserLock(name='my_lock')
            >>> user_lock.get_lock_info()
            {'locked': True, 'name': 'my_lock'}
        """
        return self.get_response(self_obj=self)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from copy import deepcopy
from inspect import isroutine
from types import GenericAlias, UnionType
from typing import Any, Self

from everysk.core.exceptions import DefaultError, FieldValueError, RequiredError

CLASS_KEY: str = '__class_path__'


class SimpleField:
    ## Private attributes
    _required_empty_values: tuple = (None, Undefined, '', [], {}, set(), frozenset())

    ## Public attributes
    # We set these attributes to Undefined so they can be set later and the field can
    # be initialized properly if it is set to be readonly, required, or has a default value None.
    attr_type: type | UnionType = Undefined
    attr_name: str = Undefined
    choices: set = Undefined
    default: Any = Undefined
    readonly: bool = Undefined
    required: bool = Undefined
    required_lazy: bool = Undefined
    empty_is_none: bool = Undefined

    ## Private methods
    def __init__(
        self,
        default: Any = None,
        *,
        choices: set | None = None,
        readonly: bool = False,
        required: bool = False,
        **kwargs,
    ) -> None:
        required_lazy = kwargs.pop('required_lazy', False)
        if required and required_lazy:
            raise FieldValueError("Required and required_lazy can't be both True.")

        # Set additional attributes from kwargs first because
        # they can change the value of the other attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.choices = set(choices) if choices else set()
        self.empty_is_none = kwargs.pop('empty_is_none', False)
        self.readonly = readonly
        self.required = required
        self.required_lazy = required_lazy

        # We use a function to clean the default value so we must set it after the other attributes
        self.default = self._clean_default(default)

        # We need to validate the default value if it is set
        # on init we do not validate by required because it can be set to Undefined or None
        if self.default is not None and self.default is not Undefined:
            self.validate(self.default)

        if readonly and (self.default is Undefined or self.default is None):
            raise RequiredError('If field is readonly, then default value is required.')

    def __eq__(self, value: object) -> bool:
        if isinstance(value, type(self)):
            # If the object is of the same type, we compare the attributes
            return self.__dict__ == value.__dict__

        return False

    def __getattr__(self, name: str) -> Any:
        # https://pythonhint.com/post/2118347356810295/avoid-pylint-warning-e1101-instance-of-has-no-member-for-class-with-dynamic-attributes
        if not isinstance(self.attr_type, UnionType):
            return getattr(self.attr_type, name)

        # We try to get the attribute from the ones that are in the tuple
        for attr_type in getattr(self.attr_type, '__args__', []):
            try:
                return getattr(attr_type, name)
            except AttributeError:
                pass

        # If no attribute was found we raise the error
        msg = f"type object '{self.attr_type}' has no attribute '{name}'."
        raise AttributeError(msg)

    def __ne__(self, value: object) -> bool:
        return not self.__eq__(value)

    def __setattr__(self, name: str, value: Any) -> None:
        # If the field is readonly, we could not change its values
        if self.readonly:
            old_value = getattr(self, name, Undefined)
            if old_value is not Undefined and old_value != value:
                self._raise_readonly_error()

        return super().__setattr__(name, value)

    def __str__(self) -> str:
        return str(self.default)

    def __repr__(self) -> str:
        cls = type(self)
        return cls.__name__

    def _clean_default(self, value: Any) -> Any:
        value = self.clean_value(value)

        if isinstance(value, (list, dict)) and not value:
            # For default values they can't be empty [] or {} - because this can cause
            # some issues with class attributes where these type can aggregate values.
            raise DefaultError('Default value cannot be a list or a dict.')

        return value

    ## Validation methods
    def _validate_attr_type(self, value: Any) -> None:  # pylint: disable=inconsistent-return-statements
        # We always accept these 3
        if value is None or value is Undefined or self.attr_type is Any:
            return

        # If attr_type is string, we check if the class name matches
        if isinstance(self.attr_type, str) and type(value).__name__ == self.attr_type:
            return

        # For subscriptable types like List, Dict, Tuple, etc.
        # We need to check if the value is a instance of the origin type
        # We not check the content of the list, dict, tuple, etc
        if isinstance(self.attr_type, GenericAlias):
            attr_type = self.attr_type.__origin__
        else:
            attr_type = self.attr_type

        if isinstance(value, attr_type):
            return

        self._raise_attr_type_error()

    def _validate_choices(self, value: Any) -> None:
        if value not in self.choices:
            self._raise_choices_error(value)

    def _validate_readonly(self, value: Any) -> None:
        # This is necessary to be able to at least assign the default value to the field
        if value != self.default:
            self._raise_readonly_error()

    def _validate_required(self, value: Any) -> None:
        if value in self._required_empty_values:
            self._raise_required_error()

    ## Raise methods
    def _raise_attr_type_error(self) -> None:
        msg = f'Key {self.attr_name} must be {self.attr_type}.'
        raise FieldValueError(msg)

    def _raise_choices_error(self, value: Any) -> None:
        msg = f"The value '{value}' for field '{self.attr_name}' must be in this list {self.choices}."
        raise FieldValueError(msg)

    def _raise_readonly_error(self) -> None:
        msg = f"The field '{self.attr_name}' value cannot be changed."
        raise FieldValueError(msg)

    def _raise_required_error(self) -> None:
        msg = f'The {self.attr_name} attribute is required.'
        raise RequiredError(msg)

    ## Public methods
    def clean_value(self, value: Any) -> Any:
        if self.empty_is_none and value == '':
            return None

        return value

    def validate(self, value: Any) -> None:
        # Required validation is always first to ensure that the field is not empty
        if not self.required_lazy and self.required:
            self._validate_required(value)

        # Run the other validations
        self._validate_attr_type(value)
        if self.choices:
            self._validate_choices(value)

        if self.readonly:
            self._validate_readonly(value)


class MetaSimpleObject(type):
    def __new__(mcs, name: str, bases: tuple, attrs: dict) -> Self:
        # Compute all attributes from the class and its bases
        attributes = {}
        # Bases are only the classes that are directly inherited from
        for cls in bases:
            # They have already been processed and have their attributes set
            attributes.update(getattr(cls, '_attributes', None) or {})

        # Add the attributes from the current class
        # We get all fields from the class annotations
        annotations = attrs.get('__annotations__', {})
        for key, value in annotations.items():
            # All annotations are types, so we need to create a SimpleField for each one
            field = SimpleField(attr_type=value, default=attrs.get(key), attr_name=key)
            # After the field is created, we need to set it in the class if it is not already set
            if key not in attrs:
                attrs[key] = field.default

            attributes[key] = field

        # We need to add the attributes that are not in the annotations
        for key in attrs.keys() - annotations.keys():
            # We need to discard python builtins and class methods/properties
            if not key.startswith('__') and not isroutine(value) and not isinstance(value, property):
                value = attrs.get(key, Undefined)
                # If the value is not a SimpleField, we create a new one
                if not isinstance(value, SimpleField):
                    value = SimpleField(attr_type=type(value), default=value, attr_name=key)
                else:
                    # If the value is a SimpleField, we need to set the attr_name
                    value.attr_name = key

                # And update the value in the attributes dictionary
                attrs[key] = value.default

                # We set the field in the attributes dictionary
                attributes[key] = value

        # Add the attributes as an attribute to the class
        attrs['_attributes'] = attributes

        return super().__new__(mcs, name, bases, attrs)

    def __setattr__(cls, name: str, value: Any) -> None:
        # if the attribute exists in the _attributes dictionary, we need to validate it
        if name in cls._attributes:
            # First we need to clean the value
            value = cls._attributes[name].clean_value(value)
            # Then we validate the value
            cls._attributes[name].validate(value)

        return super().__setattr__(name, value)


class SimpleObject(dict, metaclass=MetaSimpleObject):
    ## Private attributes
    # The _attributes is set by the metaclass when the class is created we put it here
    # to be able to access it in the instance methods and for autocomplete works.
    _attributes: dict[str, SimpleField]
    _errors: dict[str, Exception] = None
    _invalid_key_names: frozenset[str] = frozenset([])
    _need_validation: bool = True
    _silent: bool = False
    _to_dict_exclude_keys: frozenset[str] = frozenset([])
    _to_dict_key_mapping: dict[str, str] = None

    ## Init methods
    def __before_init__(self, **kwargs) -> dict:
        return kwargs

    def __init_attributes__(self) -> None:
        # Initialize attributes that are not part of the dictionary
        if self._errors is None:
            self._errors = {}

        if self._to_dict_key_mapping is None:
            self._to_dict_key_mapping = {}

    def __init_kwargs__(self, **kwargs) -> None:
        # Get all default keys from the class attributes
        # Remove the keys that are already in kwargs
        # and remove the keys that are already set in the instance
        keys = self._attributes.keys() - kwargs.keys() - self.__dict__.keys()
        kwargs.update({key: getattr(self, key, None) for key in keys})
        return kwargs

    def __init__(self, *args, **kwargs) -> None:
        # Transform args and kwargs into a dictionary to better handle them
        kwargs = dict(*args, **kwargs)
        self._silent = kwargs.pop('silent', kwargs.pop('_silent', False))

        # Call the before_init method to allow subclasses to modify kwargs
        try:
            self.__init_attributes__()
            kwargs = self.__init_kwargs__(**kwargs)
            kwargs = self.__before_init__(**kwargs)
        except Exception as error:  # pylint: disable=broad-exception-caught
            if not self._silent:
                raise error
            self._errors['before_init'] = deepcopy(error)

        # Validate the keys in kwargs
        try:
            # Remove all private keys from kwargs
            private_keys = {k for k in kwargs if self._is_invalid_key(k)}
            for key in private_keys:
                setattr(self, key, kwargs.pop(key))

            # We need to check if the values are valid before setting them
            kwargs = {k: self._clean_and_validate_value(k, v) for k, v in kwargs.items()}

            # Initialize the dictionary with the remaining keys
            super().__init__(**kwargs)

        except Exception as error:  # pylint: disable=broad-exception-caught
            if not self._silent:
                raise error
            self._errors['init'] = deepcopy(error)

        # Call the after_init method to allow subclasses to perform additional initialization
        try:
            self.__after_init__()
        except Exception as error:  # pylint: disable=broad-exception-caught
            if not self._silent:
                raise error
            self._errors['after_init'] = deepcopy(error)

    def __after_init__(self) -> None:
        pass

    ## Internal methods
    def __delattr__(self, name: str) -> None:
        if name in self:
            del self[name]
        else:
            super().__delattr__(name)

    def __getattribute__(self, name):
        # This is to keep the values only accessible through the dictionary
        if name in self:
            return self[name]

        return super().__getattribute__(name)

    def __getitem__(self, key: str) -> Any:
        if self._is_invalid_key(key):
            raise KeyError(f"Invalid key: '{key}'.")

        return super().__getitem__(key)

    def __setattr__(self, name: str, value: Any) -> None:
        # Not valid keys are set directly on the instance
        if self._is_invalid_key(name):
            value = self._clean_and_validate_value(name, value)
            super().__setattr__(name, value)
        else:
            # Public keys are set in the dictionary
            self[name] = value

    def __setitem__(self, key, value):
        if self._is_invalid_key(key):
            raise KeyError(f"Invalid key: '{key}'.")

        value = self._clean_and_validate_value(key, value)
        return super().__setitem__(key, value)

    ## Private methods
    def _clean_and_validate_value(self, key: str, value: Any) -> Any:
        if key in self._attributes:
            # First we clean the value
            value = self._attributes[key].clean_value(value)

            # Then we validate the value if needed
            if self._need_validation:
                self._attributes[key].validate(value)

        return value

    def _is_invalid_key(self, key: str) -> bool:
        return key.startswith('_') or key in self._invalid_key_names

    @classmethod
    def _get_full_doted_class_path(cls) -> str:
        return f'{cls.__module__}.{cls.__name__}'

    ## Public methods
    def copy(self) -> Self:
        return deepcopy(self)

    def fromkeys(self, keys: list, default: Any = None) -> dict:
        cls = type(self)
        return cls({key: self.get(key, getattr(self, key, default)) for key in keys})

    def get_full_doted_class_path(self) -> str:
        cls = type(self)
        return cls._get_full_doted_class_path()  # pylint: disable=protected-access

    def to_dict(self, add_class_path: bool = False, recursion: bool = False) -> dict:
        dct = {}
        for key, value in self.items():
            if key in self._to_dict_exclude_keys:
                continue

            if recursion and isinstance(value, SimpleObject):
                value = value.to_dict(add_class_path=add_class_path, recursion=recursion)

            func = getattr(self, f'_process_{key}', None)
            if callable(func):
                value = func(value)

            mapped_key = self._to_dict_key_mapping.get(key, key)
            dct[mapped_key] = value

        if add_class_path:
            dct[CLASS_KEY] = self.get_full_doted_class_path()

        return dct

    def update(self, *args, **kwargs):
        dct = dict(*args, **kwargs)
        has_invalid_keys = [key for key in dct if self._is_invalid_key(key)]
        if has_invalid_keys:
            raise KeyError(f'Invalid keys: {", ".join(has_invalid_keys)}.')

        super().update(*args, **kwargs)

    def validate_required_fields(self) -> None:
        for key, field in self._attributes.items():
            if getattr(field, 'required_lazy', False):
                field._validate_required(getattr(self, key, Undefined))  # pylint: disable=protected-access
